"""Rest client for accessing the fibaro API."""
from __future__ import annotations

import logging
from typing import Any

from requests import Response, Session
from requests.auth import HTTPBasicAuth
from requests.exceptions import JSONDecodeError

from .const import DEFAULT_TIMEOUT, HTTP_HEADERS

_LOGGER = logging.getLogger(__name__)


class RestClient:
    """Rest client for fibaro home center."""

    def __init__(
        self, url: str, username: str | None = None, password: str | None = None
    ) -> None:
        """Init"""
        self._session = Session()
        self._session.headers = HTTP_HEADERS
        self._base_url = url
        if username and password:
            self.set_auth(username, password)

    def set_auth(self, username: str, password: str) -> None:
        """Set the credentials for the fibaro home center."""
        self._session.auth = HTTPBasicAuth(username, password)

    def get(
        self, endpoint: str, json: Any | None = None, timeout: int | None = None
    ) -> Any:
        """Execute a get request."""
        current_timeout = timeout if timeout else DEFAULT_TIMEOUT
        response = self._session.get(
            f"{self._base_url}{endpoint}", json=json, timeout=current_timeout
        )

        return self._process_json_result(response)

    def post(
        self, endpoint: str, json: Any | None = None, timeout: int | None = None
    ) -> Any:
        """Execute a post request."""
        current_timeout = timeout if timeout else DEFAULT_TIMEOUT
        response = self._session.post(
            f"{self._base_url}{endpoint}", json=json, timeout=current_timeout
        )

        return self._process_json_result(response)

    def close(self) -> None:
        """Close the session."""
        self._session.close()

    def _process_json_result(self, resp: Response) -> Any:
        """Do error handling and logging on a HTTP response and covert to json."""
        _LOGGER.debug(
            '%s "%s": %s', resp.request.method, resp.request.url, resp.status_code
        )

        resp.raise_for_status()

        try:
            json = resp.json()
            _LOGGER.debug("Response: %s", json)
            return json
        except JSONDecodeError:
            _LOGGER.debug("No response")
            return None



"""Endpoint object to access the endpoint settings/info"""
from __future__ import annotations

import logging
from typing import Any

from .common.const import IGNORE_DEVICE
from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


class DeviceModel:
    """Model of a device."""

    def __init__(self, data: dict, rest_client: RestClient, api_version: int) -> None:
        """Constructor."""
        self.raw_data = data
        self._rest_client = rest_client
        self._api_version = api_version

    @property
    def fibaro_id(self) -> int:
        """Device id"""
        return int(self.raw_data.get("id", 0))

    @property
    def name(self) -> str:
        """Device name"""
        return self.raw_data.get("name")

    @property
    def parent_fibaro_id(self) -> int:
        """Id of the parent device or 0 if there is no parent."""
        return int(self.raw_data.get("parentId", 0))

    @property
    def type(self) -> str | None:
        """Device type."""
        return self.raw_data.get("type")

    @property
    def base_type(self) -> str | None:
        """Device base type."""
        return self.raw_data.get("baseType")

    @property
    def room_id(self) -> int:
        """Room id of the device or 0 if no room is assigned."""
        return int(self.raw_data.get("roomID", 0))

    @property
    def properties(self) -> dict:
        """Get the properties."""
        return self.raw_data.get("properties", {})

    @property
    def actions(self) -> dict[str, int]:
        """Get the available actions."""
        return self.raw_data.get("actions", {})

    def execute_action(self, action: str, arguments: list[Any] | None = None) -> Any:
        """Execute a device action.

        Params:
        action: name of the action to call
        arguments: list of arguments needed for the action
        """
        if action not in self.actions:
            _LOGGER.warning(
                "The device %s has no action %s. Possible actions are %s",
                self.fibaro_id,
                action,
                self.actions,
            )

        url = f"devices/{self.fibaro_id}/action/{action}"
        if arguments:
            args_prepared = {"args": arguments}
            _LOGGER.debug(
                "Execute %s for device %s with args %s.",
                action,
                self.fibaro_id,
                args_prepared,
            )
            return self._rest_client.post(url, json=args_prepared)

        _LOGGER.debug("Execute %s for device %s without args.", action, self.fibaro_id)
        return self._rest_client.post(url)

    @staticmethod
    def read_devices(rest_client: RestClient, api_version: int) -> list[DeviceModel]:
        """Returns a list of devices."""
        raw_data: list[dict] = rest_client.get("devices")

        devices: list[DeviceModel] = []
        for device in raw_data:
            if device.get("type") in IGNORE_DEVICE:
                _LOGGER.debug("Ignore device: %s", device.get("id"))
            else:
                devices.append(device)
        return [DeviceModel(data, rest_client, api_version) for data in devices]



"""Endpoint object to access the endpoint settings/info"""

from .common.rest_client import RestClient


class LoginModel:
    """Fibaro login."""

    def __init__(self, rest_client: RestClient) -> None:
        """Load the data."""
        self.raw_data: dict = rest_client.get("loginStatus")

    @property
    def is_logged_in(self) -> bool:
        """Returns the software version."""
        return self.raw_data.get("status")



"""Main class for accessing fibaro API."""

from .common.rest_client import RestClient
from .fibaro_device import DeviceModel
from .fibaro_info import InfoModel
from .fibaro_login import LoginModel
from .fibaro_room import RoomModel
from .fibaro_scene import SceneModel
from .fibaro_state_handler import FibaroStateHandler


class FibaroClient:
    """Fibaro client.

    This is the main entry point to access the fibaro API.

    Usage:
    Use set_authentication() to provide the credentials
    Use connect() to establish the connection and check if the credentials are valid
    Use any other method to access API data and actions
    """

    def __init__(self, url: str) -> None:
        """Init the fibaro client.

        The url needs to be in the format http://<HOST>/api/.
        """
        self._rest_client = RestClient(url)
        self._api_version: int = None
        self._state_handler: FibaroStateHandler = None

    def set_authentication(self, username: str, password: str) -> None:
        """Set the credentials."""
        self._rest_client.set_auth(username, password)

    def connect(self) -> bool:
        """Returns the login status.

        Returns:
        True if authenticated, False if not authenticated

        Raises:
        HTTPError: If there is a connection problem. Most important is HTTPError
        with status 403 which raised if invalid credentials are provided.
        """
        login = LoginModel(self._rest_client)

        # Read the API version as it is needed regularly
        self._api_version = self.read_info().api_version

        return login.is_logged_in

    def read_info(self) -> InfoModel:
        """Read the info endpoint from home center."""
        return InfoModel(self._rest_client)

    def read_rooms(self) -> list[RoomModel]:
        """Read the rooms endpoint from home center."""
        return RoomModel.read_rooms(self._rest_client)

    def read_scenes(self) -> list[SceneModel]:
        """Read the scenes endpoint from home center."""
        return SceneModel.read_scenes(self._rest_client, self._api_version)

    def read_devices(self) -> list[DeviceModel]:
        """Read the devices endpoint from home center."""
        return DeviceModel.read_devices(self._rest_client, self._api_version)

    def register_update_handler(self, callback: callable) -> None:
        """Register a state handler."""
        if self._state_handler:
            raise Exception("There is already a state handler registered")
        self._state_handler = FibaroStateHandler(self._rest_client, callback)

    def unregister_update_handler(self) -> None:
        """Unregister the state handler."""
        if self._state_handler:
            self._state_handler.stop()
            self._state_handler = None



"""Endpoint object to access the endpoint settings/info"""
import logging

from .common.const import API_VERSION_MATCHER
from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


class InfoModel:
    """Fibaro info."""

    def __init__(self, rest_client: RestClient) -> None:
        """Load the data."""
        self.raw_data: dict = rest_client.get("settings/info")

    @property
    def current_version(self) -> str:
        """Returns the software version."""
        return self.raw_data.get("currentVersion").get("version")

    @property
    def serial_number(self) -> str:
        """Returns the serial number of the home center."""
        return self.raw_data.get("serialNumber")

    @property
    def hc_name(self) -> str:
        """Returns the serial number of the home center."""
        return self.raw_data.get("hcName")

    @property
    def api_version(self) -> int:
        """Returns the API version. As of writing version 4 and 5 was supported.

        When the version cannot be evaluated, it will fallback to version 4.
        """
        serial_number = self.serial_number
        for item in API_VERSION_MATCHER.items():
            if serial_number.startswith(item[0]):
                _LOGGER.debug("API version %s found by pattern %s", item[1], item[0])
                return item[1]
        return 4



"""Endpoint object to access the endpoint settings/info"""

from __future__ import annotations

from .common.rest_client import RestClient


class RoomModel:
    """Model of a room."""

    def __init__(self, data: dict) -> None:
        """One room."""
        self.raw_data = data

    @property
    def fibaro_id(self) -> int:
        """Room id"""
        return self.raw_data.get("id")

    @property
    def name(self) -> str:
        """Room name"""
        return self.raw_data.get("name")

    @staticmethod
    def read_rooms(rest_client: RestClient) -> list[RoomModel]:
        """Returns a list of rooms."""
        raw_data: list = rest_client.get("rooms")
        return [RoomModel(data) for data in raw_data]



"""Endpoint object to access the endpoint settings/info"""
from __future__ import annotations

from .common.rest_client import RestClient


class SceneModel:
    """Model of a scene."""

    def __init__(self, data: dict, rest_client: RestClient, api_version: int) -> None:
        """One scne."""
        self.raw_data = data
        self._rest_client = rest_client
        self._api_version = api_version

    @property
    def fibaro_id(self) -> int:
        """Scene id"""
        return self.raw_data.get("id")

    @property
    def name(self) -> str:
        """Scene name"""
        return self.raw_data.get("name")

    @property
    def room_id(self) -> int:
        """Room id of the scene or 0 if no room is assigned."""
        return int(self.raw_data.get("roomID", 0))

    def start(self) -> None:
        """Start a scene."""
        if self._api_version == 4:
            self._send_action_v4("start")
        else:
            self._send_action_v5("execute")

    def stop(self) -> None:
        """Stop a scene."""
        if self._api_version == 4:
            self._send_action_v4("stop")
        else:
            self._send_action_v5("kill")

    def _send_action_v4(self, action: str) -> None:
        url = f"scenes/{self.fibaro_id}/action/{action}"
        self._rest_client.post(url)

    def _send_action_v5(self, action: str) -> None:
        url = f"scenes/{self.fibaro_id}/{action}"
        self._rest_client.post(url, {})

    @staticmethod
    def read_scenes(rest_client: RestClient, api_version: int) -> list[SceneModel]:
        """Returns a list of scenes."""
        raw_data: list = rest_client.get("scenes")
        return [SceneModel(data, rest_client, api_version) for data in raw_data]
